# Copyright 2018 Oliver Berger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Very simple caching pypi proxy."""
import abc
import asyncio
import hashlib
import pathlib
import re
import time
import urllib

import aiofiles
import aiohttp
import aiohttp.web
import aiohttp_jinja2
import aiotask_context
import attr
import jinja2
import pkg_resources
import structlog
import inotipy

from . import utils

log = structlog.get_logger()
routes = aiohttp.web.RouteTableDef()
re_semver = re.compile(
    r'\bv?(?:0|[1-9]\d*)'
    r'\.(?:0|[1-9]\d*)'
    r'\.(?:0|[1-9]\d*)'
    r'(?:-[\da-z-]+(?:\.[\da-z-]+)*)'
    r'?(?:\+[\da-z-]+(?:\.[\da-z-]+)*)?\b')

default_cache_timeout = 60 * 60 * 24
"""Cache timeout defaults to one day."""


@routes.get('/')
async def index(request):
    config = request.config_dict['config']
    context = {
        'projects': [],
        'principal': 'default',
    }
    response = aiohttp_jinja2.render_template(
        'simple.j2', request, context)
    return response


@routes.get(r'/release/{path:[a-f0-9/]+}/{file}', name='release')
async def serve_release(request):
    pypi_cache = request.config_dict['pypi_cache']
    cached_release = await CachedRelease.from_cached_metadata(
        cache=pypi_cache,
        path=pathlib.Path(request.match_info['path'])
    )
    headers = {
        'Content-disposition':
        f'attachment; filename={cached_release.file_name}'
    }
    return aiohttp.web.Response(
        body=cached_release.streamer.stream(),
        headers=headers
    )


@routes.get(r'/{project_name}')
@routes.get(r'/{project_name}/')
@routes.get(r'/{project_name}/{json:json}')
@routes.get(r'/{project_name}/{version:' + re_semver.pattern + r'}')
@routes.get(r'/{project_name}/{version:' + re_semver.pattern + r'}/')
@routes.get(r'/{project_name}/{version:' + re_semver.pattern + r'}/{json:json}')    # noqa
async def get_project(request):
    pypi_cache = request.config_dict['pypi_cache']
    project_name = request.match_info['project_name']
    version = request.match_info.get('version', None)
    project = CachedProject(project_name, version, cache=pypi_cache)

    try:
        metadata = await project.get_metadata()
    except MetadataNotFound:
        raise aiohttp.web.HTTPNotFound()
    except MetadataRetrievingError:
        raise aiohttp.web.HTTPInternalServerError()

    # if we have a json request, like the new pypi API
    if request.match_info.get('json', False):
        return utils.json_response(metadata)

    context = {
        'project_name': project_name,
        'metadata': metadata,
    }
    response = aiohttp_jinja2.render_template(
        'simple_package.j2', request, context)
    return response


class CacheException(Exception):
    pass


class MetadataNotFound(CacheException):
    pass


class MetadataRetrievingError(CacheException):
    pass


class Cache:        # noqa: R0903
    def __init__(self, cache_root, cache_timeout=None):
        self.cache_root = cache_root
        self.cache_timeout = cache_timeout or default_cache_timeout

    def principal_project_path(self, project_name, principal=None):
        path = self.cache_root.joinpath(
            'principals',
            'default' if principal is None else principal,
            project_name[0],
            project_name
        )
        return path

    def project(self, project_name):
        project = CachedProject(project_name, cache=self)
        return project


class Project:      # noqa: R0903
    def __init__(self, project_name, version=None):
        self.project_name = project_name
        self.version = version

    @abc.abstractmethod
    async def get_metadata(self):
        """Return the metadata for that project and version."""


class PypiProject(Project):
    api_base = 'https://pypi.org/pypi/{project_version}/json'.format

    @utils.reify
    def url(self):
        project_version = (self.project_name
                           if self.version is None
                           else '/'.join((self.project_name, self.version)))
        return self.api_base(project_version=project_version)

    async def get_metadata(self):
        client_session = aiotask_context.get('client_session')
        async with client_session.get(self.url) as r:
            if r.status == 200:
                log.info('Loading metadata from upstream', url=self.url)
                metadata = await r.json()
                return metadata

            if r.status == 404:
                log.error('Metadata not found', project=self)
                raise MetadataNotFound(self, r)

            log.error('Error while retrieving metadata', project=self)
            raise MetadataRetrievingError(self, r)


class CachedProject(PypiProject):

    """A Project is a cached pypi project."""

    def __init__(self, project_name, version=None, *, cache):
        super().__init__(project_name, version)
        self.cache = cache

    @utils.reify
    def cache_path(self):
        path = self.cache.principal_project_path(self.project_name)
        if not path.is_dir():
            path.mkdir(parents=True)
        return path

    @utils.reify
    def cached_metadata_upstream_path(self):
        """The cached version of upstream metadata."""
        name = ('pypi-metadata.json'
                if self.version is None
                else f'pypi-metadata-{self.version}.json')
        path = self.cache_path.joinpath(name)
        return path

    @utils.reify
    def cached_metadata_path(self):
        """Our transformed metadata.

        We change urls of releases here.
        """
        name = ('metadata.json'
                if self.version is None
                else f'metadata-{self.version}.json')
        path = self.cache_path.joinpath(name)
        return path

    async def get_metadata(self):
        """Either lookup in cache or query pypi and store in cache."""
        # try to find actual versions if not modified since cache_timeout
        path = self.cached_metadata_path
        if (path.is_file()
                and (time.time() - path.stat().st_mtime)
                < self.cache.cache_timeout):
            # serve this file
            async with aiofiles.open(path) as f:
                log.info('Loading metadata from cache', path=path)
                metadata = utils.json_loads(await f.read())
        else:
            # create cache
            upstream_metadata = await super().get_metadata()
            async with aiofiles.open(
                    self.cached_metadata_upstream_path, "w") as f:
                await f.write(utils.json_dumps(upstream_metadata))
            # transform upstream
            metadata = await self.transform_metadata(upstream_metadata)
            async with aiofiles.open(self.cached_metadata_path, "w") as f:
                await f.write(utils.json_dumps(metadata))
        return metadata

    async def transform_metadata(self, upstream_metadata):
        """Inject our own release URL into upstream metadata."""

        async def _release_urls(urls):
            urls = [await self.transform_upstream_url(url,
                                                      prepare_cache=True)
                    for url in urls]
            return urls

        # TODO XXX FIXME
        # noqa file a bug: SyntaxError: asynchronous comprehension outside of an asynchronous function
        releases = {
            version: await _release_urls(urls)
            for version, urls in upstream_metadata['releases'].items()
        }
        urls = [
            await self.transform_upstream_url(url)
            for url in upstream_metadata['urls']
        ]

        metadata = {
            'info': upstream_metadata['info'],
            'last_serial': upstream_metadata['last_serial'],
            'releases': releases,
            # files for this version
            'urls': urls
        }
        return metadata

    async def transform_upstream_url(self, url, prepare_cache=False):
        """Transform the upstream url and prepare cache.

        Since pypi sends multiple equal urls in metadata, we do not want to
        always prepare the cache.

        """
        cached_release = CachedRelease(cache=self.cache, url=url)
        if prepare_cache:
            # TODO prepare cache by examining cache_path for existing metadata
            await cached_release.prepare()

        transformed_url = await cached_release.transform_url()

        return transformed_url


@attr.s
class CachedRelease:
    cache = attr.ib()
    url = attr.ib(default=None)

    @utils.reify
    def upstream_url(self):
        upstream_url = self.url['url']
        return upstream_url

    @utils.reify
    def cache_path(self):
        path = self.cache.cache_root.joinpath('releases')
        return path

    @utils.reify
    def path(self):
        encoded_upstream_url = self.upstream_url.encode('utf-8')
        upstream_url_hashs = [
            hashlib.blake2s(encoded_upstream_url,                       # noqa: E1101
                            digest_size=digest_size,
                            person=b'cacheReL').hexdigest()
            for digest_size in (1, 2, hashlib.blake2s.MAX_DIGEST_SIZE)  # noqa: E1101
        ]
        path = pathlib.Path().joinpath(*upstream_url_hashs)
        return path

    @utils.reify
    def file_name(self):
        parsed_upstream_url = urllib.parse.urlparse(self.upstream_url)
        file_name = pathlib.Path(parsed_upstream_url.path).name
        return file_name

    @utils.reify
    def release_file_path(self):
        path = self.cache_path.joinpath(self.path, self.file_name)
        return path

    @utils.reify
    def release_file_path_preparing(self):
        path = self.release_file_path.with_name(
            self.release_file_path.name + '.preparing')
        return path

    async def transform_url(self):
        url = self.url.copy()
        request = aiotask_context.get('request')

        url['url'] = request.app.router['release'].url_for(
            path=str(self.path),
            file=self.file_name
        )
        return url

    def metadata_path(self, path):
        metadata_path = self.cache_path.joinpath(path, 'metadata.json')
        return metadata_path

    async def prepare(self):
        # copy url metadata to cache path
        metadata_path = self.metadata_path(self.path)
        if not metadata_path.parent.is_dir():
            metadata_path.parent.mkdir(parents=True)

        async with aiofiles.open(metadata_path, 'w') as f:
            data = utils.json_dumps(self.url)
            await f.write(data)
            log.info('Cached url metadata', path=metadata_path)

    @classmethod
    async def from_cached_metadata(cls, cache, path):
        cached_release = cls(cache)
        metadata_path = cached_release.metadata_path(path)
        async with aiofiles.open(metadata_path) as f:
            data = await f.read()
            cached_release.url = utils.json_loads(data)
        return cached_release

    @property
    def streamer(self):
        streamer = CachingStreamer(self.upstream_url, self.release_file_path)
        return streamer


class CachingStreamer:

    log = structlog.get_logger(':'.join((__module__, __qualname__)))    # noqa: E0602

    def __init__(self, url, file_path):
        self.url = url
        self.file_path = file_path

    @utils.reify
    def file_path_preparing(self):
        path = self.file_path.with_name(
            self.file_path.name + '.preparing')
        return path

    async def stream(self, force=False):
        """Stream a released version file.

        If there is a cache, we assume this release version is final and we
        stream from it. Otherwise we download the data, store and stream it.

        :param force: force recaching
        """
        if not self.file_path.is_file() or force:
            if self.file_path_preparing.is_file():
                # serve from intermediate file
                async for data in self._stream_from_intermediate():
                    self.log.debug('Stream data', size=len(data))
                    yield data
            else:
                async for data in self._stream_and_cache():
                    self.log.debug('Stream data', size=len(data))
                    yield data
        else:
            self.log.info('Serving cache', path=self.file_path)
            async with aiofiles.open(self.file_path, 'rb') as f:
                async for data in ChunkedFileIterator(f, chunk_size=2**14):
                    self.log.debug('Stream data', size=len(data))
                    yield data

    async def _stream_and_cache(self):
        """Stream data from upstream and cache them.

        The download and caching is done in the background, to prevent
        disconnecting clients from stopping it.
        """
        client_session = aiotask_context.get('client_session')
        self.log.info('Caching upstream', url=self.url, path=self.file_path)

        queue = asyncio.Queue()
        fut_finished = asyncio.Future()

        async def _stream_queue():
            while queue.qsize() or not fut_finished.done():
                data = await queue.get()
                try:
                    yield data
                finally:
                    queue.task_done()

        async def _enqueue_upstream():
            try:
                async with aiofiles.open(self.file_path_preparing, 'xb') as f:
                    async with client_session.get(self.url) as r:
                        async for data, _ in r.content.iter_chunks():
                            await f.write(data)
                            await queue.put(data)
                        fut_finished.set_result(True)
                self.file_path_preparing.rename(self.file_path)
                self.log.info('Finished download', path=self.file_path)
            except (asyncio.CancelledError, IOError, Exception) as ex:      # noqa: W0703
                fut_finished.set_exception(ex)
                # cleanup broken download
                self.log.error('Cleaning broken download',
                               path=self.file_path_preparing, error=ex)
                try:
                    self.file_path_preparing.unlink()
                except FileNotFoundError:
                    pass

        # TODO use aiojobs ??? to cancel this future graceully
        # GeneratorExit
        asyncio.ensure_future(_enqueue_upstream())
        async for data in _stream_queue():
            yield data

    async def _stream_from_intermediate(self):
        self.log.info('Stream from intermediate',
                      path=self.file_path_preparing)

        watcher = inotipy.Watcher.create()
        watcher.watch(str(self.file_path_preparing),
                      inotipy.IN.MOVE_SELF
                      | inotipy.IN.DELETE_SELF
                      | inotipy.IN.CLOSE_WRITE
                      | inotipy.IN.MODIFY
                      )
        fut_finished = asyncio.Future()
        ev_write = asyncio.Event()

        async def _wait_for_event():
            while True:
                event = await watcher.get()
                self.log.debug('File event',
                               file_event=event, watch=event.watch)
                if event.mask & inotipy.IN.MODIFY:
                    ev_write.set()
                if event.mask & inotipy.IN.DELETE_SELF:
                    fut_finished.set_exception(FileNotFoundError(event))
                    break

                if event.mask & (
                        inotipy.IN.MOVE_SELF | inotipy.IN.CLOSE_WRITE):
                    fut_finished.set_result(event)
                    break

        async with aiofiles.open(self.file_path_preparing, 'rb') as f:
            while True:
                data = await f.read()
                if data:
                    yield data
                elif fut_finished.done():
                    self.log.info('Intermediate finished',
                                  result=await fut_finished)
                    break
                else:
                    # wait for next write event
                    await ev_write.wait()
                    ev_write.clear()


class ChunkedFileIterator:

    """Iterate and yield binary data chunks from a file.

    :param file: the file
    :param chunk_size: the size of the chunks to yield

    .. code-block:: python

        async for data in ChunkedFileIterator(f):
            yield data

    """
    def __init__(self, file, chunk_size=2**14):
        self.file = file
        self.chunk_size = chunk_size

    def __aiter__(self):
        """We are our own iterator."""
        return self

    async def __anext__(self):
        """Simulate normal file iteration."""
        chunk = await self.file.read(self.chunk_size)
        if chunk:
            return chunk
        raise StopAsyncIteration


def jinja2_filter_basename(path):
    return pathlib.Path(str(path)).name


def jinja2_filter_dirname(path):
    return pathlib.Path(str(path)).parent


async def plug_me_in(app):
    await app.plugin('pypare.plugins:plugin_task_context')
    pypi_cache = Cache(
        cache_root=app.config['cache_root'],
        cache_timeout=app.config['cache_timeout'],
    )
    log.info('Using cache', cache_root=pypi_cache.cache_root,
             cache_timeout=pypi_cache.cache_timeout)

    pypi_app = aiohttp.web.Application()
    pypi_app['pypi_cache'] = pypi_cache
    pypi_app.add_routes(routes)

    template_path = pkg_resources.resource_filename(__package__, 'templates')
    aiohttp_jinja2.setup(
        pypi_app,
        loader=jinja2.FileSystemLoader(template_path),
        filters={
            'basename': jinja2_filter_basename,
            'dirname': jinja2_filter_dirname,
        },
    )
    app.add_subapp(app.config['base_path'], pypi_app)
    log.info('Added subapp', prefix=app.config['base_path'],
             resources=pypi_app.router._resources)



# Copyright 2018 Oliver Berger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common middlewares to enhance microservices."""
import asyncio

import aiohttp.web
import aiohttp.client
import aiotask_context as context


async def plugin_task_context(app):
    """Add actual config, request and a :py:obj:`ClientSession` to the task
    context."""
    loop = asyncio.get_event_loop()
    loop.set_task_factory(context.task_factory)
    client_session = aiohttp.client.ClientSession()

    async def close_client_session(app):    # noqa: W0613
        await client_session.close()
    app.on_cleanup.append(close_client_session)

    @aiohttp.web.middleware
    async def task_context_middleware(request, handler):        # noqa: W0613
        context.set('request', request)
        context.set('config', app.config)
        context.set('client_session', client_session)
        response = await handler(request)
        return response

    app.middlewares.append(task_context_middleware)



# Copyright 2018 Oliver Berger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import logging.config
import sys

import aiohttp
import structlog


LOGGING_LEVEL_NAMES = list(map(logging.getLevelName, sorted((
    logging.NOTSET, logging.DEBUG, logging.INFO,
    logging.WARN, logging.ERROR, logging.CRITICAL,
))))
DEFAULT_LOGGING_LEVEL = logging.getLevelName(logging.INFO)
CONTEXT_SETTINGS = dict(token_normalize_func=lambda x: x.upper())


def is_debug(level):
    return level == logging.getLevelName(logging.DEBUG)


class StdioToLog:

    """Delegate sys.stdout to a logger."""

    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ''

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())

    def flush(self):
        pass


class ExtractLogExtra:      # noqa: R0903

    """Extract log record attributes to structlog event_dict."""

    def __init__(self, *attrs):
        self.attrs = attrs

    def __call__(self, logger, method_name, event_dict):
        """
        Add the logger name to the event dict.
        """
        record = event_dict.get("_record")
        for attr_name in self.attrs:
            if hasattr(record, attr_name):
                attr = getattr(record, attr_name)
                event_dict[attr_name] = attr
        return event_dict


def setup_logging(*, tty=sys.stdout.isatty(),
                  level=logging.DEBUG, capture_warnings=True,
                  redirect_print=False, json_indent=None):
    """Set up structured logging for logstash.

    :param tty: if True renders colored logs
    :param level: the log level to be applied
    :param capture_warnings: redirect warnings to the logger
    :param redirect_print: use the logger to redirect printed messages
    """
    # normalize level
    if isinstance(level, str):
        level = logging.getLevelName(level.upper())

    renderer = (structlog.dev.ConsoleRenderer()
                if tty else
                structlog.processors.JSONRenderer(indent=json_indent,
                                                  sort_keys=True)
                )
    timestamper = structlog.processors.TimeStamper(fmt="ISO", utc=True)
    pre_chain = [
        # Add the log level and a timestamp to the event_dict if the log entry
        # is not from structlog.
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.processors.format_exc_info,
        ExtractLogExtra('spec', 'url', 'mimetype', 'has_body', 'swagger_yaml',
                        'method', 'path', 'operation_id', 'data'),
        timestamper,
    ]

    logging.config.dictConfig({
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "plain": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": renderer,
                "foreign_pre_chain": pre_chain,
            },
            "colored": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": renderer,
                "foreign_pre_chain": pre_chain,
            },
        },
        "handlers": {
            "default": {
                # "level": "DEBUG",
                "class": "logging.StreamHandler",
                "formatter": "colored",
            },
            # "file": {
            #     # "level": "DEBUG",
            #     "class": "logging.handlers.WatchedFileHandler",
            #     "filename": "test.log",
            #     "formatter": "plain",
            # },
        },
        "loggers": {
            "": {
                "handlers": ["default"],
                "level": level,
                "propagate": True,
            },
        }
    })
    logging.captureWarnings(capture_warnings)
    processors = [
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.PositionalArgumentsFormatter(),
        timestamper,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ]

    structlog.configure(
        processors=processors,
        context_class=structlog.threadlocal.wrap_dict(dict),
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    if redirect_print:
        # redirect stdio print
        print_log = structlog.get_logger('print')
        sys.stderr = StdioToLog(print_log)
        sys.stdout = StdioToLog(print_log)

    # log uncaught exceptions
    sys.excepthook = uncaught_exception


def uncaught_exception(ex_type, ex_value, tb):  # noqa: C0103
    log_ = structlog.get_logger('sys.excepthook')
    log_.critical(event='uncaught exception', exc_info=(ex_type, ex_value, tb))


def merge_override_maps(maps):
    """Merge all maps and remove emptied keys."""
    merged = {}
    for m in maps:
        merged.update(m)

    remove_keys = {k for k, v in merged.items() if not v}

    for k in remove_keys:
        del merged[k]

    return merged


class AccessLogger(aiohttp.abc.AbstractAccessLogger):   # noqa: R0903
    def log(self, request, response, time):
        log = structlog.get_logger()
        log.info('Access',
                 remote=request.remote,
                 method=request.method,
                 path=request.path,
                 time=time,
                 status=response.status)



# Copyright 2018 Oliver Berger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import importlib
import inspect
import itertools
import sys

import aiohttp.web
import structlog

from . import logging


class Config(dict):
    defaults = {}
    """Some config defaults."""

    sanitizer = {}
    """Some converter to sanitize config values."""

    def __init__(self, config, **kwargs):
        sanitizer = self.sanitizer
        sane_config = (
            (k, sanitizer[k](v) if k in sanitizer else v)
            for k, v in itertools.chain(config.items(), kwargs.items())
        )

        super().__init__(sane_config)
        self.set_defaults(self.defaults)
        self._loaded_plugins = set()

    def set_defaults(self, config):
        """Applies default values from config.

        :param config: another :py:obj:`dict`
        """
        for key, value in config.items():
            self.setdefault(key, value)


class AioHttpConfig(Config):
    defaults = {
        'host': '0.0.0.0',
        'port': 8080,
        'plugins_function_name': 'plug_me_in',
        'plugins': [],
        'debug': True,
    }
    """Some config defaults."""

    sanitizer = {
        'plugins': list,
        'port': int,
        'debug': bool
    }
    """Some converter to sanitize config values."""

    def create_app(self):
        """Create the aiohttp app."""
        app = aiohttp.web.Application()
        app._debug = self['debug']
        # we hook in our self
        app['config'] = self
        app.config = self
        app.plugin = functools.partial(self.plugin, app)
        return app

    def prepare_app(self, app):
        self.hook_plugins(app)

    async def plugin(self, app, name):
        plugin = resolve_dotted_name(name)

        if inspect.ismodule(plugin):
            # apply default name
            plugin = getattr(plugin, self.get('plugins_function_name'))

        if not inspect.iscoroutinefunction(plugin):
            raise ValueError(f'{plugin.__name__} must be an async function.')

        log = structlog.get_logger()
        if plugin in self._loaded_plugins:
            log.warning('Plugin already loaded',
                        plugin=f'{plugin.__module__}.{plugin.__qualname__}')
            return

        try:
            return await plugin(app)
        finally:
            log.info('Loaded plugin',
                     plugin=f'{plugin.__module__}.{plugin.__qualname__}')
            self._loaded_plugins.add(plugin)

    def hook_plugins(self, app):
        """Actually load all async plugins."""
        async def _apply_plugins():
            for name in self['plugins']:
                await self.plugin(app, name)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(_apply_plugins())

    def run(self):
        app = self.create_app()
        self.prepare_app(app)

        log = structlog.get_logger()
        log.info('Starting aiohttp', host=self['host'], port=self['port'])
        aiohttp.web.run_app(
            app,
            host=self['host'],
            port=self['port'],
            access_log_class=logging.AccessLogger,
            print=None
        )


def resolve_dotted_name(name):
    """Use pkg_resources style dotted name to resolve a name.

    Modules are cached in sys.modules
    """
    part = ':'
    module_name, _, attr_name = name.partition(part)
    if part in attr_name:
        raise ValueError(f'Invalid name: {name}')

    if module_name in sys.modules:
        resolved = sys.modules[module_name]
    else:
        spec = importlib.util.find_spec(module_name)
        resolved = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(resolved)
        sys.modules[resolved.__name__] = resolved

    if attr_name:
        resolved = getattr(resolved, attr_name)

    return resolved



# Copyright 2018 Oliver Berger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import json
import pathlib

import aiohttp.web
import structlog
import yarl

json_adapter_registry = {}
"""A mapping of types to json adapter functions."""


def json_adapter(json_cls, adapter=None):
    """Register an adapter function to transform an obj into something json
    serializable."""
    if adapter is None:
        return lambda func: json_adapter(json_cls, func)

    if json_cls not in json_adapter_registry:
        json_adapter_registry[json_cls] = adapter
    else:
        log = structlog.get_logger()
        log.warning('JSON adapter for this type already registered',
                    type=json_cls, adapter=json_adapter_registry[json_cls])

    return adapter


def find_json_adapter(json_obj):
    """Find the appropriate adapter based on the MRO of the instance type."""
    for json_obj_cls in json_obj.__class__.__mro__:
        adapter = json_adapter_registry.get(json_obj_cls)
        return adapter


def json_default_handler(obj):
    """Try to convert an object to a json renderable type.

    If that object has a ``__json__`` method we call it first.
    After all we try to adapt that result via registered json adapters.
    """
    if hasattr(obj, '__json__') and callable(obj.__json__):
        result = obj.__json__()
    else:
        result = obj

    # try to adapt
    adapter = find_json_adapter(obj)
    if adapter:
        result = adapter(obj)

    return result


json_loads = json.loads
json_dumps = functools.partial(json.dumps, default=json_default_handler)
json_response = functools.partial(aiohttp.web.json_response, dumps=json_dumps)


@json_adapter(pathlib.Path)
def adapt_path(obj):
    return str(obj)


@json_adapter(yarl.URL)
def adapt_url(obj):
    return str(obj)


class reify:        # noqa: R0903
    def __init__(self, wrapped):
        self.wrapped = wrapped
        functools.update_wrapper(self, wrapped)

    def __get__(self, inst, cls=None):
        if inst is None:
            return self
        val = self.wrapped(inst)
        setattr(inst, self.wrapped.__name__, val)
        return val



# Copyright 2018 Oliver Berger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import importlib
import pathlib

import click
import structlog

from .. import config, logging, __version__


def find_loop_specs():
    """Just find specs for common loops."""
    module_specs = (
        (module_name, importlib.util.find_spec(module_name))
        for module_name in ('asyncio', 'uvloop', 'tokio')
    )
    available_specs = {
        module_name: spec for module_name, spec in module_specs
        if spec is not None
    }
    return available_specs


def set_loop_policy(event_loop):
    log = structlog.get_logger()
    if event_loop == 'uvloop':
        try:
            import uvloop
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
            log.info("Using uvloop event loop policy")

        except ImportError:
            log.warning("uvloop is not available.")

    elif event_loop == 'tokio':
        try:
            import tokio
            asyncio.set_event_loop_policy(tokio.EventLoopPolicy())
            log.info("Using tokio event loop policy")

        except ImportError:
            log.warning("tokio is not available.")


@click.group(context_settings=logging.CONTEXT_SETTINGS)
@click.option('--log-level', default=logging.DEFAULT_LOGGING_LEVEL,
              show_default=True,
              type=click.Choice(logging.LOGGING_LEVEL_NAMES),
              help='The logging level.')
@click.option('event_loop', '--loop', default='asyncio',
              show_default=True,
              type=click.Choice(find_loop_specs().keys()),
              help='Use a different loop policy.')
@click.version_option(__version__)
@click.pass_obj
def cli(obj, log_level, event_loop):
    obj['debug'] = logging.is_debug(log_level)
    logging.setup_logging(level=log_level)
    set_loop_policy(event_loop)


class Path(click.Path):
    def convert(self, value, param, ctx):
        value = super().convert(value, param, ctx)
        value = pathlib.Path(value).expanduser().resolve()
        return value


@cli.command('pypi')
@click.option('-p', '--port', default=3141, type=int,
              show_default=True,
              help='The port to run the server')
@click.option('-h', '--host', default='0.0.0.0',
              show_default=True,
              help='The server host IP.')
@click.option('base_path', '-b', '--base-path', type=click.Path(),
              default='/pypi', show_default=True,
              help='The base path for this application.')
@click.option('cache_root', '-c', '--cache-root',
              type=Path(file_okay=False, dir_okay=True, writable=True),
              default='~/.cache/pypare', show_default=True,
              help='The cache directory, where files are stored.',
              )
@click.option('cache_timeout', '--cache-timeout', type=int,
              default=60 * 60 * 24, show_default=True,
              help='The age of metatdata, when it will be refreshed.')
@click.option('plugins', '--plugin', multiple=True, type=list,
              help='A plugin in pkg_resources notation to load.')
@click.pass_obj
def cli_cache(obj, **pypi_config):
    """Run a simple pypi caching proxy."""
    conf = config.AioHttpConfig(pypi_config, debug=obj['debug'])
    conf['plugins'].append('pypare.pypi')
    conf.run()


# @cli.command('resolve')
# def cli_resolve():
#     pass


def main():
    cli(obj={}, auto_envvar_prefix='PYPARE')    # noqa: E1123



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['pypare', 'pypare.scripts']

package_data = \
{'': ['*'], 'pypare': ['templates/*']}

install_requires = \
['aiofile>=1.4,<2.0',
 'aiofiles>=0.4.0,<0.5.0',
 'aiohttp-jinja2>=1.1,<2.0',
 'aiohttp>=3.4,<4.0',
 'aiotask_context>=0.6.0,<0.7.0',
 'click>=7.0,<8.0',
 'colorama>=0.3.9,<0.4.0',
 'inotipy>=0.1.1,<0.2.0',
 'structlog>=18.2,<19.0']

entry_points = \
{'console_scripts': ['pypare = pypare.scripts:main']}

setup_kwargs = {
    'name': 'pypare',
    'version': '0.2.3',
    'description': 'A very simple pypi cache',
    'long_description': 'pypare\n======\n\nA very simple pypi cache.\n\nfeatures\n^^^^^^^^\n\n- uses `aiohttp`_, `aiofiles`_, `inotipy`_\n\n- queries metadata via pypi JSON API\n\n- filesystem is the database\n\n- serve releases while downloading\n\n\n.. _`aiohttp`: http://aiohttp.readthedocs.io/\n.. _`aiofiles`: https://pypi.org/project/aiofiles/\n.. _`inotipy`: https://github.com/ldo/inotipy\n\ntodo\n^^^^\n\n- private channels with user, groups and permissions\n\n- use `python-libaio`_ for file stuff\n\n- nice ui\n\n.. _`python-libaio`: https://github.com/vpelletier/python-libaio\n\n\nrunning the cache\n^^^^^^^^^^^^^^^^^\n\n.. code-block::\n\n    # pypare --help\n    Usage: pypare [OPTIONS] COMMAND [ARGS]...\n\n    Options:\n      --log-level [NOTSET|DEBUG|INFO|WARNING|ERROR|CRITICAL]\n                                      The logging level.  [default: INFO]\n      --loop [asyncio|uvloop]         Use a different loop policy.  [default:\n                                      asyncio]\n      --version                       Show the version and exit.\n      --help                          Show this message and exit.\n\n    Commands:\n      pypi  Run a simple pypi caching proxy.\n\n\n.. code-block::\n\n    # pypare pypi --help\n    Usage: pypare pypi [OPTIONS]\n\n      Run a simple pypi caching proxy.\n\n    Options:\n      -p, --port INTEGER          The port to run the server  [default: 3141]\n      -h, --host TEXT             The server host IP.  [default: 0.0.0.0]\n      -b, --base-path PATH        The base path for this application.  [default:\n                                  /pypi]\n      -c, --cache-root DIRECTORY  The cache directory, where files are stored.\n                                  [default: ~/.cache/pypare]\n      --cache-timeout INTEGER     The age of metatdata, when it will be refreshed.\n                                  [default: 86400]\n      --plugin LIST               A plugin in pkg_resources notation to load.\n      --help                      Show this message and exit.\n\nRun from virtual environment:\n\n.. code-block:: bash\n\n   pip install pypare\n\n   pypare pypi --cache-dir /tmp/pypi-data\n\n\nRun in docker:\n\n.. code-block:: bash\n\n   docker run -it diefans/pypare:latest pypi\n',
    'author': 'Oliver Berger',
    'author_email': 'oliver@digitalarchitekt.de',
    'url': 'https://github.com/diefans/pypare',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
